#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <string.h>
#include <time.h>
#include "util.h"
#include "align.h"

void insert(char c, int ref_pos, char **genome, int **map, int *len)
{
  **genome = c;
  (*genome)++;
  **map = ref_pos;
  (*map)++;
  (*len)++;
}

void delete(char **genome, int **map, int *len)
{
  (*genome)--;
  (*map)--;
  (*len)--;
}

void substitute(char **genome, int **map, int *len)
{
  delete(genome, map, len);
  char s = 'A';
  if (**genome == 'A') {
    s = 'T';
  }
  insert(s, **map, genome, map, len);
}

bool add_errors(char *a,
                int a_length,
                char **b,
                int *b_length,
                char **c,
                int *c_length,
                int **b_mapping,
                int **c_mapping,
                int *errors,
                unsigned int seed,
                float chance)
{
  *b = malloc(sizeof(char) * ((a_length * 2) + 1));
  *c = malloc(sizeof(char) * ((a_length * 2) + 1));
  *b_mapping = malloc(sizeof(int) * a_length * 2);
  *c_mapping = malloc(sizeof(int) * a_length * 2);
  if (*b == NULL || *c == NULL || *b_mapping == NULL || *c_mapping == NULL) {
    fprintf(stderr, "memory allocation in add_errors failed\n");
    return false;
  }

  char *b_head = *b;
  char *c_head = *c;
  int *b_mapping_head = *b_mapping;
  int *c_mapping_head = *c_mapping;

  srand(seed);
  for (int i = 0; i < a_length; i++) {
    insert(a[i], i, b, b_mapping, b_length);
    insert(a[i], i, c, c_mapping, c_length);

    int both = (rand() / (double)RAND_MAX) > 0.5;
    double r = rand() / (double)RAND_MAX;
    if (r < chance) {
      insert(a[i], i, b, b_mapping, b_length);
      if (both) {
        insert(a[i], i, c, c_mapping, c_length);
      }
    } else if (r < (2 * chance)) {
      substitute(b, b_mapping, b_length);
      if (both) {
        substitute(c, c_mapping, c_length);
      }
    } else if (r < (3 * chance)) {
      delete(b, b_mapping, b_length);
      if (both) {
        delete(c, c_mapping, c_length);
      }
    }
    if (r < (3 * chance)) {
      (*errors)++;
    }
  }
  **b = '\0';
  **c = '\0';

  *b = realloc(b_head, sizeof(char) * ((*b_length) + 1));
  *c = realloc(c_head, sizeof(char) * ((*c_length) + 1));
  *b_mapping = realloc(b_mapping_head, sizeof(int) * (*b_length));
  *c_mapping = realloc(c_mapping_head, sizeof(int) * (*c_length));
  if (*b == NULL || *c == NULL || *b_mapping == NULL || *c_mapping == NULL) {
    fprintf(stderr, "memory allocation in add_errors failed\n");
    return false;
  }

  return true;
}

void benchmark_align(int times,
                     char *A,
                     int A_length,
                     char *B1,
                     int B1_length,
                     char *B2,
                     int B2_length,
                     int *m1, int *m2,
                     int e)
{
  double total = 0;
  long k = -1;
  for (int i = 0; i < times; i++) {
    time_t start = time(NULL);
    k = align(A, A_length,
              B1, B1_length,
              B2, B2_length,
              A_length,
              m1, m2);
    total += difftime(time(NULL), start);
    assert(k <= e);
  }

  printf("align | size: %d, avg_time: %f, errors: %d, distance: %ld\n",
         A_length, total / times, e, k);
}

int read_genome(FILE *fp, char **out)
{
  int s = 1024;
  char *o = malloc(sizeof(char) * s);
  
  int i = 0;
  int c = fgetc(fp);
  while (c != EOF) {
    if (c != '\n') {
      if (!(i < s - 1)) {
        s = s * 2;
        o = realloc(o, sizeof(char) * s);
      }
      o[i++] = (char)c;
    }
    c = fgetc(fp);
  }

  o[i] = '\0';
  o = realloc(o, sizeof(char) * (i + 1));
  (*out) = o;
  return i;
}

int main(int argc, char *argv[]) {
  if (argc < 2) {
    fprintf(stderr, "Usage: benchmark input.txt [times]\n");
    return EXIT_FAILURE;
  }

  int times = 1;
  if (argc > 2) {
    if (sscanf(argv[2], "%d", &times) < 1 ||
        times < 1) {
      fprintf(stderr, "invalid times: %s\n", argv[2]);
      return EXIT_FAILURE;
    }
  }

  FILE *f = fopen(argv[1], "r");
  if (f == NULL) {
    fprintf(stderr, "cannot open file %s\n", argv[1]);
    return EXIT_FAILURE;
  }

  char *a;
  int l = read_genome(f, &a);
  fclose(f);

  char *B1;
  char *B2;
  int *m1;
  int *m2;
  int l1;
  int l2;
  int e = 0;

  bool s = add_errors(a, l,
                      &B1, &l1,
                      &B2, &l2,
                      &m1, &m2, &e,
                      1, 0.01 / 3);

  if (s) {
    benchmark_align(times, a, l, B1, l1, B2, l2, m1, m2, e);
    return EXIT_SUCCESS;
  } else {
    return EXIT_FAILURE;
  }
}
